"""
utils.py

Various utilities used in fragmenting molecules - retrieving information from molecules
Identifying small parts of molecules, and tools for minor manipulations of SMILES or molecules
"""
import ast
import os

import pandas as pd
from rdkit import Chem  # pylint:disable=import-error
from rdkit.Chem import (  # pylint:disable=import-error # search for rdScaffoldAttachment points * to remove; pylint:disable=import-error; rdDetermineBonds,
    AllChem,
    rdDetermineBonds,
    rdqueries,
)


def get_molecules_atomicnum(molecule):
    """Given molecule object, get list of atomic numbers."""
    # a
    atom_num_list = []
    for atom in molecule.GetAtoms():
        atom_num_list.append(atom.GetAtomicNum())
    return atom_num_list


def get_molecules_atomsinrings(molecule):
    """Given molecule object, get Boolean list of if atoms are in a ring."""
    is_in_ring_list = []
    for atom in molecule.GetAtoms():
        is_in_ring_list.append(atom.IsInRing())
    return is_in_ring_list


def trim_placeholders(rwmol):
    """Given Chem.RWmol, remove atoms with atomic number 0."""
    qa = rdqueries.AtomNumEqualsQueryAtom(0)  # define query for atomic number 0
    if len(rwmol.GetAtomsMatchingQuery(qa)) > 0:  # if there are matches
        query_match = rwmol.GetAtomsMatchingQuery(qa)
        rm_at_idx = []
        for atom in query_match:  # identify atoms to be removed
            rm_at_idx.append(atom.GetIdx())
            # remove starting from highest number so upcoming indices not affected
        rm_at_idx_sort = sorted(rm_at_idx, reverse=True)
        # e.g. need to remove 3 and 5, if don't do this and you remove 3,
        # then the 5 you want to remove is now 4, and you'll remove wrong atom
        for idx in rm_at_idx_sort:  # remove atoms
            rwmol.RemoveAtom(idx)
    return rwmol


def mol_with_atom_index(mol):
    """Label a molecule with map numbers for order atoms are in molecule.

    Useful for tracking atoms after fragmenting

    Args:
        mol: an rdkit molecule object

    Returns:
        rdkit molecule object with AtomMapNum set so that atom 0
        has map number 1, atom 1 has map number 2...etc

    """
    # from https://www.rdkit.org/docs/Cookbook.html
    for atom in mol.GetAtoms():
        if atom.GetAtomicNum() != 0:
            atom.SetAtomMapNum(atom.GetIdx() + 1)
    return mol


def _get_charge_from_cml(cml_file):
    with open(cml_file, encoding="utf-8") as file:
        for line in file:
            if "formalCharge" in line:
                split_line = line.split(" ")
                for word in split_line:
                    if "formalCharge" in word:
                        charge = int(
                            word.replace("formalCharge=", "")
                            .replace(">\n", "")
                            .replace('"', "")
                        )
                        break
    return charge


def mol_from_xyzfile(xyz_file: str, cml_file):
    """Attempt to create a molecule from an xyz file by automatically determining bond orders"""
    # raw_mol = Chem.MolFromXYZFile('DUDE_67368827_adrb2_decoys_C19H25N3O4_CIR.xyz')
    # mol = Chem.Mol(raw_mol)
    # rdDetermineBonds.DetermineConnectivity(mol)
    # rdDetermineBonds.DetermineBondOrders(mol)
    mol = Chem.Mol(Chem.MolFromXYZFile(xyz_file))
    try:
        rdDetermineBonds.DetermineBonds(mol, charge=_get_charge_from_cml(cml_file))
    except:  # pylint:disable=bare-except
        if os.path.isfile("error_log.txt"):
            with open("error_log.txt", "a", encoding="utf-8") as er_file:
                er_file.write(
                    f"Could not determine bond orders from xyz for {cml_file}\n"
                )
        else:
            with open("error_log.txt", "w", encoding="utf-8") as er_file:
                er_file.write(f"Could not determine bond orders for {cml_file}\n")
        return None

    atomic_symbols = []
    xyz_coordinates = []
    ats_read = 0
    num_atoms = mol.GetNumAtoms()
    with open(xyz_file, encoding="utf-8") as file:
        for line_number, line in enumerate(file):
            if ats_read < num_atoms and line_number > 1:
                ats_read += 1
                atomic_symbol, x, y, z = line.split()[:4]
                atomic_symbols.append(atomic_symbol)
                xyz_coordinates.append([float(x), float(y), float(z)])
            elif ats_read == num_atoms:
                break
    return {
        "Molecule": mol_with_atom_index(mol),
        "xyz_pos": xyz_coordinates,
        "atomic_symbols": atomic_symbols,
    }


# from https://github.com/rdkit/rdkit/issues/2413
# conf = m.GetConformer()


# in principal, you should check that the atoms match
# for i in range(m.GetNumAtoms()):
#     print(i)
#     x,y,z = xyz_coordinates[i]
#     conf.SetAtomPosition(i,Point3D(x,y,z))


def get_cml_atom_types(cml_file):
    """Extract atom types from cml file

    Args:
        cml_file: cml file name

    Returns:
        Data Frame column whose elements are tuples of
        atom types as defined by Retrievium. (atom number, type, valence)

    """
    n_atl = 0
    type_list = []
    idx_list = []
    with open(cml_file, encoding="utf-8") as file:
        for line in file:
            if "atomTypeList" in line:
                n_atl += 1
                if n_atl == 2:
                    break
            elif n_atl == 1:
                split_line = line.split()
                idx_list.append(
                    int(split_line[1].split("=")[1].replace('"', "").replace("a", ""))
                    - 1
                )
                at_label = split_line[2].split("=")[1].replace('"', "")
                at_type = int(split_line[3].split("=")[1].replace('"', ""))
                at_valence = int(
                    split_line[4].split("=")[1].split("/")[0].replace('"', "")
                )
                type_list.append((at_label, at_type, at_valence))
    temp_frame = pd.DataFrame(list(zip(idx_list, type_list)), columns=["idx", "type"])
    temp_frame.sort_values(by="idx", inplace=True)
    return temp_frame["type"]


def add_cml_atoms_bonds(el_list, bond_list):
    """create a molecule from cml file, building it one atom at a time then
    adding in bond orders.

    Note: Bond orders from Retrievium cml are problematic at times.

    Recommended construction is to build atoms and connectivity from cml file
    Then assign bond orders based on template smiles also found in cml file"""
    flag = 1
    for atom in el_list:
        if flag:
            mol = Chem.MolFromSmiles(atom)
            rwmol = Chem.RWMol(mol)
            rwmol.BeginBatchEdit()
            flag = 0
        else:
            rwmol.AddAtom(Chem.Atom(atom))
    # mw.AddBond(6,7,Chem.BondType.SINGLE)
    for bond in bond_list:
        if bond[2] == "S":
            rwmol.AddBond(bond[0] - 1, bond[1] - 1, Chem.BondType.SINGLE)
        elif bond[2] == "D":
            rwmol.AddBond(bond[0] - 1, bond[1] - 1, Chem.BondType.DOUBLE)
        elif bond[2] == "T":
            rwmol.AddBond(bond[0] - 1, bond[1] - 1, Chem.BondType.TRIPLE)
        elif bond[2] == "A":
            rwmol.AddBond(bond[0] - 1, bond[1] - 1, Chem.BondType.AROMATIC)
    rwmol.CommitBatchEdit()
    return rwmol


def _add_cml_single_atoms_bonds(el_list, bond_list):
    """Build a mol from list of atoms and bonds, one atom at a time
    Bonds assigned are only single bonds.

    Adjust bonds after with _modAssignBondOrdersFromTemplate"""
    rwmol = Chem.RWMol(Chem.Mol())
    for atom in el_list:
        rwmol.AddAtom(Chem.Atom(atom))
    # mw.AddBond(6,7,Chem.BondType.SINGLE)
    for bond in bond_list:
        rwmol.AddBond(bond[0] - 1, bond[1] - 1, Chem.BondType.SINGLE)
    rwmol.CommitBatchEdit()
    return rwmol


def smiles_from_cml(cml_file, smile_tag="retrievium:inputSMILES"):
    """Finds the Retreivium  SMILES in a cml file with a given label

    Args:
        cml_file: cml file name
        smile_tag: the label fo the SMILEs in the cml file. Defaults to input SMILEs

    Returns:
        string of the input SMILES code tagged in the file as retrievium:inputSMILES

    Note:
        Must be used on .cml files from the Retrievium database https://retrievium.ca

    """
    flag = 0
    with open(cml_file, encoding="utf-8") as file:
        for line in file:
            if smile_tag in line:
                flag = 1
            elif flag == 1:
                smile = line.split(">")[1].split("<")[0]
                break
    return smile


def mol_from_cml(cml_file, input_type="cmlfile"):
    """Creates a molecule from a cml file and returns atoms, xyz and types

    Builds molecule one atom at a time connected by single bonds
    Then determines bond orders by mapping to a template smiles in the cml
    Finally, updates property cache, initializes ring info, and sanitizes mol
    If no match between SMILEs and connectivity, returns None and writes
    error to file

    Args:
        cml_file: cml file name or dictionary containing cml data
        input_type: cmlfile or cmldict. Use cmlfile if just raw cml file, use cmldict if dictionary

    Returns:
        list with elements: [Molecule, list of atom symbols in molecule,
        list of xyz coords of atoms in molecule,
        list of atom types of atoms in molecule]

    Note:
        list order matches mol numbering in cml

    """

    if input_type == "cmlfile":
        xyz_coords, at_types, bond_list, el_list, _ = data_from_cml(cml_file, True)
        smile = smiles_from_cml(cml_file)
    elif input_type == "cmldict":
        xyz_coords = cml_file["geom"]
        at_types = cml_file["atom_types"]
        bond_list = cml_file["bonds"]
        el_list = cml_file["labels"]
        # charge = cml_file["charge"]
        smile = cml_file["smiles"]
    rwmol = _add_cml_single_atoms_bonds(el_list, bond_list)
    for atom in rwmol.GetAtoms():
        atom.SetNoImplicit(True)

    rwmol2 = Chem.RemoveHs(rwmol, implicitOnly=True, updateExplicitCount=False)
    template = AllChem.MolFromSmiles(smile)
    bond_mol = _modAssignBondOrdersFromTemplate(template, rwmol2, cml_file)
    # need rings for aromaticity check
    # if os.path.isfile('error_log.txt'):
    #     er_file = open('error_log.txt','a')
    #     er_file.write(f'Could not sanitize {cml_file}\n')
    #     er_file.close()
    # else:
    #     er_file = open('error_log.txt','w')
    #     er_file.write(f'Could not sanitize {cml_file}\n')
    #     er_file.close()
    if bond_mol:
        bond_mol.UpdatePropertyCache()
        Chem.GetSymmSSSR(bond_mol)
        try:
            Chem.SanitizeMol(bond_mol)
        except:  # pylint:disable=bare-except
            _write_error(f"Could not sanitize {cml_file}\n")
            return [None, None, None, None]
        return [mol_with_atom_index(bond_mol), el_list, xyz_coords, at_types]
        # if os.path.isfile('error_log.txt'):
        #     er_file = open('error_log.txt','a')
        #     er_file.write(f'No match between template smiles and connected geom for {cml_file}\n')
        #     er_file.close()
        # else:
        #     er_file = open('error_log.txt','w')
        #     er_file.write(f'No match between template smiles and connected geom for {cml_file}\n')
        #     er_file.close()
    return [None, None, None, None]


def _write_error(errormessage):
    """writes errormessage to errorlog.txt"""
    if os.path.isfile("error_log.txt"):
        with open("error_log.txt", "a", encoding="utf-8") as er_file:
            er_file.write(errormessage)
    else:
        with open("error_log.txt", "w", encoding="utf-8") as er_file:
            er_file.write(errormessage)


def _modAssignBondOrdersFromTemplate(refmol, mol, cml_file):
    """This is originally from RDKit AllChem module.
      Modified here by Kevin Lefrancois-Gagnon(KLG) to disallow implicit hydrogens on all
      molcule objects used. This corresponds to the 4 for loops after creation of mol
      objects.
      Also Returns None for no match instead of value error, for use on large number of
      molecules, a single error won't stop the whole thing
      All other code is unmodified from the original.
      This was required by KLG for the following reason:
        using the original AssignBondOrdersFromTemplate on a molecule with
        only explicit hydrogens with only connectivity, no  bond types other than
        single bonds. When a charged atom like nitrogen was present, extra H were being added
        For example when a part was expected as C-[NH2+]-C, the following was seen
        C-[NH2+]([H])([H])-C, essentially doubling the expected # hydrogens
        These could not be deleted by Chem.RemoveHs(mol,implicitOnly=True) as for
        some reason these were explicit. The reason for why these were added is unclear

        Beyond the obviously problematic extra hydrogens, it also produced an error case
        when the later part of the code would clean up bond orders, like for some ring ketones
        in a ring labeled aromatic by the SMILEs.
        Adjusting implicit Hs to be disallowed for all molecules and their copies at the start
        resulted in the normal expected outputs.
    Resume original documentation:
      assigns bond orders to a molecule based on the
        bond orders in a template molecule

        Arguments
          - refmol: the template molecule
          - mol: the molecule to assign bond orders to

        An example, start by generating a template from a SMILES
        and read in the PDB structure of the molecule

        >>> import os
        >>> from rdkit.Chem import AllChem
        >>> template = AllChem.MolFromSmiles("CN1C(=NC(C1=O)(c2ccccc2)c3ccccc3)N")
        >>> mol = AllChem.MolFromPDBFile(os.path.join(RDConfig.RDCodeDir, 'Chem', 'test_data', '4DJU_lig.pdb'))
        >>> len([1 for b in template.GetBonds() if b.GetBondTypeAsDouble() == 1.0])
        8
        >>> len([1 for b in mol.GetBonds() if b.GetBondTypeAsDouble() == 1.0])
        22

        Now assign the bond orders based on the template molecule

        >>> newMol = AllChem.AssignBondOrdersFromTemplate(template, mol)
        >>> len([1 for b in newMol.GetBonds() if b.GetBondTypeAsDouble() == 1.0])
        8

        Note that the template molecule should have no explicit hydrogens
        else the algorithm will fail.

        It also works if there are different formal charges (this was github issue 235):

        >>> template=AllChem.MolFromSmiles('CN(C)C(=O)Cc1ccc2c(c1)NC(=O)c3ccc(cc3N2)c4ccc(c(c4)OC)[N+](=O)[O-]')
        >>> mol = AllChem.MolFromMolFile(os.path.join(RDConfig.RDCodeDir, 'Chem', 'test_data', '4FTR_lig.mol'))
        >>> AllChem.MolToSmiles(mol)
        'COC1CC(C2CCC3C(O)NC4CC(CC(O)N(C)C)CCC4NC3C2)CCC1N(O)O'
        >>> newMol = AllChem.AssignBondOrdersFromTemplate(template, mol)
        >>> AllChem.MolToSmiles(newMol)
        'COc1cc(-c2ccc3c(c2)Nc2ccc(CC(=O)N(C)C)cc2NC3=O)ccc1[N+](=O)[O-]'

    """
    # pylint:disable=too-many-branches
    refmol = Chem.AddHs(refmol)
    refmol2 = Chem.Mol(refmol)
    refmol2 = Chem.AddHs(refmol2)
    mol2 = Chem.Mol(mol)
    for atom in mol.GetAtoms():
        atom.SetNoImplicit(True)
    for atom in mol.GetAtoms():
        atom.SetNoImplicit(True)
    for atom in refmol2.GetAtoms():
        atom.SetNoImplicit(True)
    for atom in refmol.GetAtoms():
        atom.SetNoImplicit(True)
    # do the molecules match already?
    matching = mol2.GetSubstructMatch(refmol2)
    if not matching:  # no, they don't match
        # check if bonds of mol are SINGLE Chem.rdchem.BondType.DOUBLE
        for b in mol2.GetBonds():
            if b.GetBondType() != Chem.rdchem.BondType.SINGLE:
                b.SetBondType(Chem.rdchem.BondType.SINGLE)
                b.SetIsAromatic(False)
        # set the bonds of mol to SINGLE
        for b in refmol2.GetBonds():
            b.SetBondType(Chem.rdchem.BondType.SINGLE)
            b.SetIsAromatic(False)
        # set atom charges to zero;
        for a in refmol2.GetAtoms():
            a.SetFormalCharge(0)
        for a in mol2.GetAtoms():
            a.SetFormalCharge(0)

        matching = mol2.GetSubstructMatches(refmol2, uniquify=False)
        # do the molecules match now?
        if matching:
            #   if len(matching) > 1:
            #     logger.warning(msg="More than one matching pattern found - picking one")
            matching = matching[0]
            # apply matching: set bond properties
            for b in refmol.GetBonds():
                atom1 = matching[b.GetBeginAtomIdx()]
                atom2 = matching[b.GetEndAtomIdx()]
                b2 = mol2.GetBondBetweenAtoms(atom1, atom2)
                b2.SetBondType(b.GetBondType())
                b2.SetIsAromatic(b.GetIsAromatic())
            # apply matching: set atom properties
            for a in refmol.GetAtoms():
                a2 = mol2.GetAtomWithIdx(matching[a.GetIdx()])
                a2.SetHybridization(a.GetHybridization())
                a2.SetIsAromatic(a.GetIsAromatic())
                a2.SetNumExplicitHs(a.GetNumExplicitHs())
                a2.SetFormalCharge(a.GetFormalCharge())
            try:
                Chem.SanitizeMol(mol2)
            except:  # pylint:disable=bare-except
                er_message = f"Could not sanitize {cml_file}\n"
                # smile = smiles_from_cml(cml_file)
                er_message += _at_num_er(refmol2, mol2)
                # smimol = Chem.MolFromSmiles()
                _write_error(er_message)
                return None
            if hasattr(mol2, "__sssAtoms"):  # pylint:disable=protected-access
                mol2.__sssAtoms = None  # pylint:disable=protected-access # we don't want all bonds highlighted
        else:
            er_message = (
                f"No match between template smiles and connected geom for {cml_file}\n"
            )
            er_message += _at_num_er(refmol2, mol2)
            _write_error(er_message)
            return None
    return mol2


def _at_num_er(refmol2, mol2):
    """check numbering of atoms"""
    er_message = ""
    smi_num = []
    smi_lab = []
    symb_list = []
    num_list = []
    for atom in mol2.GetAtoms():
        if atom.GetSymbol() not in symb_list:
            symb_list.append(atom.GetSymbol())
            num_list.append(1)
        else:
            num_list[symb_list.index(atom.GetSymbol())] += 1
    for atom in refmol2.GetAtoms():
        if atom.GetSymbol() not in smi_lab:
            smi_lab.append(atom.GetSymbol())
            smi_num.append(1)
        else:
            smi_num[smi_lab.index(atom.GetSymbol())] += 1
    for i, num in enumerate(num_list):
        j = smi_lab.index(symb_list[i])
        if smi_num[j] != num:
            er_message += f"Expected {smi_num[j]} {smi_lab[j]} from SMILES, observed {num_list[i]} in xyz\n"
    return er_message


def data_from_cml(cml_file, bonds=False):
    """Gets symbols, xyz coords, bonds and charge of a mol from cml file

    Args:
        cml_file: .cml filename

    Returns:
        list with relevant data from cml file.
        Elements in order are: molecular geometry, atom types,
        list of bonds, list of elements, charge


    Note:
        This is designed to parse files specifically from the Retrievium database
        https://retrievium.ca

        the SMILEs extracted is labelled with tag retrievium:inputSMILES

        The geometry extracted is from the third atomArray block in the .cml file

    Example Usage:
        >>> utils.data_from_cml(cml_file)

    """
    # pylint:disable=too-many-locals
    # pylint:disable=too-many-branches
    # pylint:disable=too-many-statements
    num_atom_array = 0
    geom_list = []
    n_atl = 0
    n_bary = 0
    type_list = []
    idx_list = []
    bond_list = []
    el_list = []
    with open(cml_file, encoding="utf-8") as file:
        for line in file:
            if "formalCharge" in line:
                split_line = line.split(" ")
                for word in split_line:
                    if "formalCharge" in word:
                        charge = int(
                            word.replace("formalCharge=", "")
                            .replace(">\n", "")
                            .replace('"', "")
                        )
                        continue
            if "atomArray" in line:
                num_atom_array += 1
                if num_atom_array == 5:
                    continue
            if num_atom_array == 5:
                quote_split = line.split('"')
                # maybe only do el_list with bonds?
                el_list.append(quote_split[3])
                x_split = quote_split[5]
                y_split = quote_split[7]
                z_split = quote_split[9]
                geom_list.append(
                    [
                        float(ast.literal_eval(x_split)),
                        float(ast.literal_eval(y_split)),
                        ast.literal_eval(z_split),
                    ]
                )
            elif num_atom_array == 6:
                if bonds:
                    if "bondArray" in line:
                        n_bary += 1
                        if n_bary in [1, 2]:
                            continue
                    elif n_bary == 1:
                        split_line = line.split()
                        at_1 = int(split_line[1].split('"')[1].replace("a", ""))
                        at_2 = int(split_line[2].replace('"', "").replace("a", ""))
                        b_ord = split_line[4].split('"')[1]
                        bond_list.append((at_1, at_2, b_ord))
                if "atomTypeList" in line:
                    n_atl += 1
                    if n_atl == 1:
                        continue
                    if n_atl == 2:
                        break
                elif n_atl == 1:
                    split_line = line.split()
                    idx_list.append(
                        int(
                            split_line[1]
                            .split("=")[1]
                            .replace('"', "")
                            .replace("a", "")
                        )
                        - 1
                    )
                    at_label = split_line[2].split("=")[1].replace('"', "")
                    at_type = int(split_line[3].split("=")[1].replace('"', ""))
                    at_valence = int(
                        split_line[4].split("=")[1].split("/")[0].replace('"', "")
                    )
                    type_list.append((at_label, at_type, at_valence))
    temp_frame = pd.DataFrame(list(zip(idx_list, type_list)), columns=["idx", "type"])
    temp_frame.sort_values(by="idx", inplace=True)
    if bonds:
        return [geom_list, list(temp_frame["type"]), bond_list, el_list, charge]
    return [geom_list, list(temp_frame["type"])]


def all_data_from_cml(data):
    """Gets symbols, xyz coords, bonds and charge of a mol from cml file

    Args:
        data: lines of a .cml file

    Returns:
        dictionary with relevant data from cml file.
        Keys included are 'geom', 'atom_types', 'bonds',
        'labels', 'charge', 'multiplicity', 'smiles'

    Note:
        Not used in :attr:`group_decomposition.fragfunctions.identify_connected_fragments`
        This is used in an AiiDA workflow employing this package

        This is designed to parse files specifically from the Retrievium database
        https://retrievium.ca

        the SMILEs extracted is labelled with tag retrievium:inputSMILES

        The geometry extracted is from the third atomArray block in the .cml file

    Example Usage:
        >>> utils.all_data_from_cml(cml_file)

    """
    # pylint:disable=too-many-locals
    # pylint:disable=too-many-branches
    # pylint:disable=too-many-statements
    num_atom_array = 0
    geom_list = []
    n_atl = 0
    n_bary = 0
    type_list = []
    idx_list = []
    bond_list = []
    el_list = []
    smi_flag = 0
    for line in data:
        if "formalCharge" in line:
            split_line = line.split(" ")
            for word in split_line:
                if "spinMultiplicity" in word:
                    multiplicity = int(
                        word.replace("spinMultiplicity=", "").replace('"', "")
                    )
                if "formalCharge" in word:
                    charge = int(
                        word.replace("formalCharge=", "")
                        .replace('"', "")
                        .replace(">", "")
                        .replace("\n", "")
                    )
                    continue
        if "retrievium:inputSMILES" in line:
            smi_flag = 1
        elif smi_flag == 1:
            smile = line.split(">")[1].split("<")[0]
            smi_flag = 2
        if "atomArray" in line:
            num_atom_array += 1
            if num_atom_array == 5:
                continue
        if num_atom_array == 5:
            quote_split = line.split('"')
            # maybe only do el_list with bonds?
            el_list.append(quote_split[3])
            x_split = quote_split[5]
            y_split = quote_split[7]
            z_split = quote_split[9]
            geom_list.append(
                [
                    float(ast.literal_eval(x_split)),
                    float(ast.literal_eval(y_split)),
                    float(ast.literal_eval(z_split)),
                ]
            )
        elif num_atom_array == 6:
            if "bondArray" in line:
                n_bary += 1
                if n_bary in [1, 2]:
                    continue
            elif n_bary == 1:
                split_line = line.split()
                at_1 = int(split_line[1].split('"')[1].replace("a", ""))
                at_2 = int(split_line[2].replace('"', "").replace("a", ""))
                b_ord = split_line[4].split('"')[1]
                bond_list.append((at_1, at_2, b_ord))
            if "atomTypeList" in line:
                n_atl += 1
                if n_atl == 1:
                    continue
                if n_atl == 2:
                    break
            elif n_atl == 1:
                split_line = line.split()
                idx_list.append(
                    int(split_line[1].split("=")[1].replace('"', "").replace("a", ""))
                    - 1
                )
                at_label = split_line[2].split("=")[1].replace('"', "")
                at_type = int(split_line[3].split("=")[1].replace('"', ""))
                at_valence = int(
                    split_line[4].split("=")[1].split("/")[0].replace('"', "")
                )
                type_list.append((at_label, at_type, at_valence))
    temp_frame = pd.DataFrame(list(zip(idx_list, type_list)), columns=["idx", "type"])
    temp_frame.sort_values(by="idx", inplace=True)

    return {
        "geom": geom_list,
        "atom_types": list(temp_frame["type"]),
        "bonds": bond_list,
        "labels": el_list,
        "charge": charge,
        "multiplicity": multiplicity,
        "smiles": smile,  # pylint:disable=used-before-assignment
    }


def xyz_list_to_str(xyz_list):
    """Convert 2d array to string [[a,b,c],[d,e,f]] -> a, b, c\n d, e, f\n"""
    outstr = ""
    for row in xyz_list:
        for at in row:
            outstr += str(at)
            outstr += ", "
        outstr += "\n"
    return outstr


def list_to_str(lst):
    """Convert a list to string. [a,b,c] -> a b c"""
    out_str = ""
    for it in lst:
        out_str += str(it) + " "
    return out_str


def mol_from_molfile(mol_file):
    """takes mol_file and returns mol wth atom numbers the same

    Modified for mol file structure from retrievium
    from stackexchange
    https://mattermodeling.stackexchange.com/questions/7234/how-to-input-3d-coordinates-from-xyz-file-and-connectivity-from-smiles-in-rdkit

    Args:
        mol_file: name of .mol file

    Returns:
        mol with AtomMapNumbers labeled by :attr:`group_decomposition.utils.mol_with_atom_index`

    """  # pylint:disable=line-too-long
    m = Chem.MolFromMolFile(mol_file, removeHs=False)
    if not m:
        raise ValueError(f"""Problem creating molecule from {mol_file}""")
    # this assumes whatever program you use doesn't re-order atoms
    #  .. which is usually a safe assumption
    #  .. so we don't bother tracking atoms
    m = mol_with_atom_index(m)
    atomic_symbols = []
    xyz_coordinates = []
    ats_read = 0
    num_atoms = m.GetNumAtoms()
    with open(mol_file, encoding="utf-8") as file:
        for line_number, line in enumerate(file):
            if ats_read < num_atoms and line_number > 3:
                ats_read += 1
                x, y, z, atomic_symbol = line.split()[:4]
                atomic_symbols.append(atomic_symbol)
                xyz_coordinates.append([float(x), float(y), float(z)])
            elif ats_read == num_atoms:
                break
    return {"Molecule": m, "xyz_pos": xyz_coordinates, "atomic_symbols": atomic_symbols}


def xyz_from_cml(cml_file):
    """Extract xyz coordinates from cml file

    Args:
        cml_file: cml file name

    Returns:
        list of length 3 lists containing a molecule's xyz coordinates

    Note:
        Must be used on .cml files from the Retrievium database https://retrievium.ca

    """
    num_atom_array = 0
    geom_list = []
    with open(cml_file, encoding="utf-8") as file:
        for line in file:
            if "atomArray" in line:
                num_atom_array += 1
                if num_atom_array == 5:
                    continue
            if num_atom_array == 5:
                space_split = line.split()
                x_split = space_split[3].split("=")
                y_split = space_split[4].split("=")
                z_split = space_split[5].split("=")
                geom_list.append(
                    [
                        float(ast.literal_eval(x_split[1])),
                        float(ast.literal_eval(y_split[1])),
                        float(ast.literal_eval(z_split[1])),
                    ]
                )
            elif num_atom_array == 6:
                break
    return geom_list


def get_canonical_molecule(smile: str):
    """Ensures that molecule numbering is consistent with creating molecule from canonical
    SMILES for consistency."""
    mol = Chem.MolFromSmiles(smile)
    if mol:
        mol_smi = Chem.MolToSmiles(mol)  # molsmi is canonical SMILES
    else:
        raise ValueError(
            f"""{smile} is not a valid SMILES code or
                         rdkit cannot construct a molecule from it"""
        )
    # create canonical molecule numbering from canonical SMILES
    return Chem.MolFromSmiles(mol_smi)


def link_molecules(mol_1: Chem.Mol, mol_2: Chem.Mol, dl_1: int, dl_2: int):
    """Given two mols, each with dummy atoms that have dummyAtomLabels, link the molecules between
    the dummy atoms specified by labels dl_1 and dl_2

    Modified from https://www.oloren.ai/blog/add_rgroup.html
    Written by David Huang, Oloren AI, modified by Kevin Lefrancois-Gagnon

    Args:
        mol_1: Chem.Mol object
        mol_2: Chem.Mol object
        dl_1: the isotope of the dummy atom in mol_1 which will be replaced by mol_2
        dl_2: the isotope of the dummy atom in mol_2 which will be replaced by mol_1

    Returns:
        Chem.Mol object with mol_1 and mol_2 linked where dl_1 and dl_2 were"""
    # pylint:disable=too-many-locals
    # Loop over atoms until there are no wildcard atoms
    # Find wildcard atom if available, otherwise exit
    # We use the isotope here are FragmentOnBonds labels the dummy atoms by changing their isotope
    a = None
    for a_ in mol_1.GetAtoms():
        if a_.GetAtomicNum() == 0 and a_.GetIsotope() == dl_1:
            a = a_
            break
    if not a:
        raise ValueError(
            f"""Input molecule mol_1 does not have atom with dummy label {dl_1}"""
        )
    b = None
    for b_ in mol_2.GetAtoms():
        if b_.GetAtomicNum() == 0 and b_.GetIsotope() == dl_2:
            b = b_
            break
    if not b:
        raise ValueError(
            f"""Input molecule mol_1 does not have atom with dummy label {dl_2}"""
        )
    # Set wildcard atoms to having AtomMapNum 1000 for tracking
    a.SetAtomMapNum(1000)
    b.SetAtomMapNum(1000)
    # Put group and base molecule together and make it editable
    m = Chem.CombineMols(mol_1, mol_2)
    m = Chem.RWMol(m)
    # Find using tracking number the atoms to merge in new molecule
    a1 = None
    a2 = None
    for at in m.GetAtoms():
        if at.GetAtomMapNum() == 1000:
            if a1 is None:
                a1 = at
            else:
                a2 = at
    # Find atoms to bind together based on atoms to merge
    b1 = a1.GetBonds()[0]
    start = (
        b1.GetBeginAtomIdx()
        if b1.GetEndAtomIdx() == a1.GetIdx()
        else b1.GetEndAtomIdx()
    )

    b2 = a2.GetBonds()[0]
    end = (
        b2.GetBeginAtomIdx()
        if b2.GetEndAtomIdx() == a2.GetIdx()
        else b2.GetEndAtomIdx()
    )

    # Add the connection and remove original wildcard atoms
    m.AddBond(start, end, order=Chem.rdchem.BondType.SINGLE)
    m.RemoveAtom(a1.GetIdx())
    m.RemoveAtom(a2.GetIdx())

    return m


# def copy_molecule(molecule):
#     """create a copy of molecule object in new object(not pointer)"""
#     # see link https://sourceforge.net/p/rdkit/mailman/message/33652439/
#     return Chem.Mol(molecule)


def _clean_smile(trim_smi):
    """remove leftover junk from smiles when atom deleted."""
    trim_smi = trim_smi.replace("[*H]", "*")
    trim_smi = trim_smi.replace("[*H3]", "*")
    trim_smi = trim_smi.replace("[*H2]", "*")
    trim_smi = trim_smi.replace("[*H+]", "*")
    trim_smi = trim_smi.replace("[*H3+]", "*")
    trim_smi = trim_smi.replace("[*H2+]", "*")
    trim_smi = trim_smi.replace("[*H-]", "*")
    trim_smi = trim_smi.replace("[*H3-]", "*")
    trim_smi = trim_smi.replace("[*H2-]", "*")
    return trim_smi



"""
fragfunctions module

code used to generate fragments of molecules from SMILES code and analyze them

Main functions to call are:
identify_connected_fragments - takes one molecule SMILES, returns fragments with connections
count_uniques - takes output from above, removes attachments and counts unique fragments
count_groups_in_set - takes list of SMILES and counts unique fragments on set
"""
# pylint:disable=too-many-lines
import math
import os
import re
from collections import Counter

import numpy as np  # for arrays in fragment identification
import pandas as pd  # lots of work with data frames
from rdkit import Chem  # pylint:disable=import-error
from rdkit.Chem import AllChem, PandasTools, rdqueries  # pylint:disable=import-error

from group_decomposition import utils

_num_bonds_broken = 1

_H_BOND_LENGTHS = {
    # from Gaussview default cleaned bond length
    "C": 1.07,
    "O": 0.96,
    "N": 1.00,
    "F": 0.88,
    "Cl": 1.29,
    "B": 1.18,
    "Al": 1.55,
    "Si": 1.47,
    "P": 1.35,
    "S": 1.31,
}


def _generate_fragment_frame(fragment_smiles):
    """Given list of SMILES Generate output frame with SMILES codes and molecules for unique
    fragments."""
    frag_frame = pd.DataFrame(set(fragment_smiles), columns=["Smiles"])
    PandasTools.AddMoleculeColumnToFrame(
        frag_frame, "Smiles", "Molecule", includeFingerprints=True
    )
    frag_frame.drop(frag_frame.index[frag_frame["Smiles"] == "*"].tolist())
    return frag_frame


def _add_xyz_coords(frag_frame):
    """Given frag_frame with molecules, add xyz coordinates form MM94 optimization to it."""
    xyz_block_list = []
    query = rdqueries.AtomNumEqualsQueryAtom(0)
    for mol in frag_frame["Molecule"]:
        h_mol_rw = Chem.RWMol(mol)  # Change type of molecule object
        h_mol_rw = Chem.AddHs(h_mol_rw)  # Add hydrogens
        zero_at = h_mol_rw.GetAtomsMatchingQuery(query)  # Replace placeholder * with At
        for atom in zero_at:
            h_mol_rw.GetAtomWithIdx(atom.GetIdx()).SetAtomicNum(85)
        AllChem.EmbedMolecule(h_mol_rw)
        AllChem.MMFFOptimizeMolecule(h_mol_rw)  # Optimize with MMFF94
        xyz_block_list.append(
            AllChem.rdmolfiles.MolToXYZBlock(h_mol_rw)  # pylint:disable=no-member
        )  # Store xyz coordinates
    frag_frame["xyz"] = xyz_block_list
    return frag_frame


def _add_number_attachements(frag_frame):
    """Add number of attachments column to frag_frame, counting number of *."""
    attach_list = []
    for molecule in frag_frame["Molecule"]:
        attach = 0
        atoms = molecule.GetAtoms()
        for atom in atoms:
            if atom.GetAtomicNum() == 0:
                attach += 1
        attach_list.append(attach)
    frag_frame["numAttachments"] = attach_list
    return frag_frame


def _fragment_molecule(
    mol_list, patt, exld_ring=False, drop_parent=True, recombine_mono=False
):
    """Break molecules into fragments based on fragmenting pattern"""
    global _num_bonds_broken  # pylint:disable=global-statement
    out_mols = []
    pat_mol = Chem.MolFromSmarts(patt)
    # print(mol_list)
    for mol in mol_list:
        if mol.HasSubstructMatch(pat_mol):
            if exld_ring and mol.GetRingInfo().NumRings() > 0:
                continue
            bonds_at_idx = mol.GetSubstructMatches(pat_mol)
            bonds_to_break = [
                mol.GetBondBetweenAtoms(x[0], x[1]).GetIdx() for x in bonds_at_idx
            ]
            labels = [
                [x, x]
                for x in range(_num_bonds_broken, _num_bonds_broken + len(bonds_at_idx))
            ]
            _num_bonds_broken += len(bonds_at_idx)
            frag_mols_comb = Chem.FragmentOnBonds(
                mol, bondIndices=bonds_to_break, dummyLabels=labels
            )
            frag_smis = Chem.MolToSmiles(frag_mols_comb).split(".")
            # print(frag_smis)
            frag_mols = [Chem.MolFromSmiles(x) for x in frag_smis]
            if recombine_mono:
                frag_mols = _recombine_monoatomic(frag_mols)

            if not drop_parent:
                out_mols = out_mols + [mol]
            out_mols = out_mols + frag_mols
        else:
            out_mols = out_mols + [mol]
    return out_mols


def _recombine_monoatomic(frag_mols):
    """Take fragments and recombine"""
    # pylint:disable=too-many-nested-blocks
    for m in frag_mols:
        link = None
        if m.GetNumHeavyAtoms() == 1:
            dummy_atoms = [x for x in m.GetAtoms() if x.GetAtomicNum() == 0]
            dummy_atom_iso = [x.GetIsotope() for x in dummy_atoms]
            dummy_match = []
            matched_iso = []
            for i, dum in enumerate(dummy_atom_iso):
                for mo in frag_mols:
                    if mo != m:
                        for at in mo.GetAtoms():
                            if at.GetAtomicNum() == 0 and at.GetIsotope() == dum:
                                dummy_match.append(mo)
                                matched_iso.append(dum)
            for i, iso in enumerate(matched_iso):
                if i == 0:
                    link = utils.link_molecules(m, dummy_match[i], iso, iso)
                else:
                    link = utils.link_molecules(link, dummy_match[i], iso, iso)
        if link:
            frag_mols.append(link)
            frag_mols.remove(m)
            for mo in dummy_match:
                frag_mols.remove(mo)
    return frag_mols


# def get_num_placeholders(mol):
#     """Find number of * in a molecule"""
#     n_p = 0
#     for atom in mol.GetAtoms():
#         if atom.GetAtomicNum() == 0:
#             n_p += 1
#     return n_p


def generate_molecule_fragments(
    mol,
    patt: str = "[$([C;X4;!R]):1]-[$([R,!$([C;X4]);!#0;!#9;!#17;!#35;!#1]):2]",
    drop_parent: bool = False,
    recombine_mono: bool = True,
):
    r"""Fragment a molecule into constituent groups

    The molecule is first broken along ring-nonring single bonds,
    then single bonds to atoms double bonded to ring (e.g. C-N={ring}),
    then breaking based on the pattern provided.



    Args:
        mol: the rdkit molecule object to be fragmented
        patt: SMARTs string that matches the bonds to be broken
             after ring-non-ring are separated
             Defaults to breaking alkyl-non-alkyl bonds. See notes
        drop_parent: if False, do include the parent structure in
            the third break. Defaults to breaking identifying alkyl chains.
            If True, do not include the parents
        recombine_mono: If True (default), will recombine separated one heavy
             atom groups with the chains they are broken from in the last step

    Returns:
        list[mol]: a list of rdkit molecules generated by fragmenting

    Note:
        Bonds broken are labelled by integers. Breaking FragA-FragB-FragC will result in FragA-1 \*:1-FragB-\*:2, \*:2-FragC

        In this way fragments can be rejoined by recombining matching integers.
        The integers are added to the molecule as Isotopes

        In the last step, halides are not separated from the alkyl groups by default,
        and one-heavy atom groups are rejoined to acyclic portions of the molecule

    """
    global _num_bonds_broken  # pylint:disable=global-statement
    _num_bonds_broken = 1
    first_break = _fragment_molecule([mol], patt="[!#0;R:1]-!@[!#0;!#1:2]")
    second_break = _fragment_molecule(
        first_break, patt="[$([!#0;!R]=[!#0;R]):1]-[!#0;!#1;!R:2]"
    )
    third_break = _fragment_molecule(
        second_break,
        patt,
        exld_ring=True,
        drop_parent=drop_parent,
        recombine_mono=recombine_mono,
    )
    return third_break


def identify_connected_fragments(
    inp: str,
    keep_only_children: bool = True,
    bb_patt: str = "[$([C;X4;!R]):1]-[$([R,!$([C;X4]);!#0;!#9;!#17;!#35;!#1]):2]",
    input_type: str = "smile",
    cml_file: str = "",
    include_parent: bool = False,
    aiida: bool = False,
) -> pd.DataFrame:
    r"""
    Given Smiles string, identify fragments in the molecule:

    Break all ring-non-ring atom single bonds.
    For atoms double bonded to rings, break their single bonds.
    For non-ring fragments, separate those into alkyl chains and hetero/double bonded atoms.
    (similar to Ertl functional groups)

    Args:
        input: a string containing either a smiles, .xyz, .mol or .cml filename for a given molecule
            update input_type below to match provided input
        keep_only_children: boolean, if True, when a group is broken down into its components
            remove the parent group from output. If False, parent group is retained
        bb_patt: string of SMARTS pattern for bonds to be broken in side chains and linkers
            defaults to cleaving sp3 carbon-((ring OR not sp3 carbon) AND not-placeholder/halogen/H)
        input_type: 'smile' if SMILES code or 'molfile' if .mol file, or 'xyzfile' if .xyz file, or 'cmlfile' if .cml file
            Note: xyz file REQUIRES cml_file to be provided as well
        cml_file: defaults to none, can be the cml file corresponding to the input .mol file
        include_parent: If True, include column in output frame repeating parent molecule
            intended use for True when merging multiple molecule fragment frames but need to retain a parent molecule object
        aiida: if True, format output to be able to be used in aiida database.
            That is, no molecule objects

    Returns:
        DataFrame with columms 'Smiles', 'Molecule', 'numAttachments' and 'xyz'
        Containing, fragment smiles, fragment Chem.Molecule object, number of \* placeholders,
        and rough xyz coordinates for the fragment is \* were At

    Note:
        Each bond breaking, connectivity is maintained through dummy atom labels.
        e.g. C-N -> C-[1*] N-[1*] - reattaching via the matching labels would reassemble the molecule

        currently will break apart a functional group if contains a ring-non-ring single bond.
        e.g. ring N-nonring S=O -> ring N-[1*] nonring S=O-[1*]

    """
    # pylint:disable=too-many-arguments
    # pylint:disable=too-many-branches
    if aiida:
        inp = inp.value
        keep_only_children = keep_only_children.value
        bb_patt = bb_patt.value
        input_type = input_type.value
        if cml_file:
            cml_file = cml_file.value
        include_parent = include_parent.value

    # ensure smiles is canonical so writing and reading the smiles will result in same number
    # ordering of atoms
    if input_type == "smile":
        mol = utils.get_canonical_molecule(inp)
        xyz_coords = []
    elif input_type in ["cmlfile", "cmldict"]:
        mol, atomic_symb, xyz_coords, atom_types = utils.mol_from_cml(
            inp, input_type=input_type
        )
        if mol is None:
            return None
    elif input_type == "molfile":
        # use coordinates in cml file provided if able, else use xyz from mol file
        mol_dict = utils.mol_from_molfile(inp)
        mol, atomic_symb = mol_dict["Molecule"], mol_dict["atomic_symbols"]
        if cml_file:
            xyz_coords, atom_types, _, _, _ = utils.data_from_cml(cml_file)
            # atom_types = utils.get_cml_atom_types(cml_file)
        else:
            xyz_coords = mol_dict["xyz_pos"]
    # elif input_type == "xyzfile":
    #     if not cml_file:
    #         raise ValueError("No cml file provided, expected one for xyz input type")
    #     mol_dict = utils.mol_from_xyzfile(xyz_file=inp, cml_file=cml_file)
    #     if mol_dict is None:
    #         return None
    #     mol, atomic_symb, xyz_coords = (
    #         mol_dict["Molecule"],
    #         mol_dict["atomic_symbols"],
    #         mol_dict["xyz_pos"],
    #     )
    #     atom_types = utils.get_cml_atom_types(cml_file)
    else:
        raise ValueError(
            f"""{input_type} should either be molfile, xyzfile, cmlfile, or a smile string"""
        )

    mol_frags = generate_molecule_fragments(
        mol, patt=bb_patt, drop_parent=keep_only_children
    )
    # initialize the output data frame
    frag_frame = _generate_fragment_frame([Chem.MolToSmiles(x) for x in mol_frags])
    # add hydrogens and xyz coordinates resulting from MMFF94 opt, changing placeholders to At
    # frag_frame = _add_xyz_coords(frag_frame)
    # count number of placeholders in each fragment - it is the number of places it is attached
    frag_frame = _add_number_attachements(frag_frame)
    if input_type in [
        "molfile",
        "xyzfile",
        "cmlfile",
        "cmldict",
    ]:  # clear map labels and add xyz coordinates that are available
        frag_frame = _add_frag_comp(frag_frame, mol)
        frag_frame["Smiles"] = frag_frame["Smiles"].map(
            _clear_map_number
        )  # (lambda x: _clear_map_number(x))
        frag_frame["xyz"] = frag_frame["Atoms"].map(
            lambda x: _add_rtr_xyz(x, xyz_coords)
        )
        frag_frame["Labels"] = frag_frame["Atoms"].map(
            lambda x: _add_rtr_label(x, atomic_symb)
        )
        frag_frame["Molecule"] = frag_frame["Molecule"].map(
            lambda x: _clear_map_number(x, "mol")
        )
        if cml_file or input_type == "cmlfile" or input_type == "cmldict":
            frag_frame["atom_types"] = frag_frame["Atoms"].map(
                lambda x: _add_rtr_type(x, atom_types)
            )
    if include_parent and not aiida:
        mol = _clear_map_number(mol, "mol")
        frag_frame["Parent"] = [mol] * len(frag_frame.index)
    if aiida:
        frag_frame = frag_frame.drop("Molecule", axis=1)
    return frag_frame


def _add_rtr_label(at_num_list, atomic_symb):
    out_list = []
    for atom in at_num_list:
        out_list.append(atomic_symb[atom - 1])
    return out_list


def _add_rtr_type(at_num_list, atom_types):
    out_list = []
    for atom in at_num_list:
        out_list.append(atom_types[atom - 1])
    return out_list


def _add_rtr_xyz(at_num_list, xyz_coords):
    """Construct string of atomSymbol x y z format, for use to map to frag_frame"""
    out_list = []
    for atom in at_num_list:
        out_list.append(xyz_coords[atom - 1])
    return out_list


def _clear_map_number(mol_input, ret_type="smi"):
    """Given str or Chem.Molecule input, remove atomMapnumbers from atoms and return
    smiles (ret_type='smi') or molecule object (ret_type = 'mol')"""
    if isinstance(mol_input, str):
        mol = Chem.MolFromSmiles(mol_input)
    else:
        mol = mol_input
    if not mol:
        raise ValueError(
            f"""Could not construct mol from {mol_input} in output frame"""
        )
    for atom in mol.GetAtoms():
        atom.ClearProp("molAtomMapNumber")
    if ret_type == "smi":
        return Chem.MolToSmiles(mol)
    if ret_type == "mol":
        return mol
    raise ValueError(f"""Invalid ret_type, expected smi or mol, got {ret_type}""")


def _get_atlabels_in_frag(molecule):
    """For a given molecule, extract list of atom map number for all non-H atoms"""
    # H's not included because the molecule object typically won't have explicit H
    out_list = []
    for atom in molecule.GetAtoms():
        if atom.GetAtomicNum() != 0:
            out_list.append(int(atom.GetProp("molAtomMapNumber")))
    return out_list


def _add_frag_comp(frag_frame, mol):
    """Given frag_frame and mol, add Atoms col to frag_frame with indices of atoms starting at 1"""
    # create list of lists of indices of atoms in each fragment
    frag_atoms = []
    for frag_mol in frag_frame["Molecule"]:
        frag_atoms.append(_get_atlabels_in_frag(frag_mol))
    # iterate over atoms, adding in hydrogens to the fragment since the above won't include H
    for atom in mol.GetAtoms():
        if atom.GetAtomicNum() == 1:
            neigh_idx = atom.GetNeighbors()[0].GetIdx() + 1
            for i, frag in enumerate(frag_atoms):
                if neigh_idx in frag:
                    frag_atoms[i].append(atom.GetIdx() + 1)
                    break  # only need to get here once - H has only one bond
    frag_frame["Atoms"] = frag_atoms
    return frag_frame


def count_uniques(
    frag_frame: pd.DataFrame, drop_attachments: bool = False, uni_smi_type: bool = False
) -> pd.DataFrame:
    r"""Identify unique fragments in a frame and count the number of times they occur

    Given frag_frame output from :attr:`group_decomposition.fragfunctions.identify_connected_fragments`,
    remove dummy atom labels(and placeholders entirely if drop_attachments=True), then count unique fragments
    using SMILES to identify unique fragments

    Should also work on frames from :attr:`group_decomposition.fragfunctions.merge_uniques` or
    :attr:`group_decomposition.fragfunctions.count_groups_in_set` as well

    Args:
        frag_frame: frame resulting from :attr:`group_decomposition.fragfunctions.identify_connected_fragments` typically,
            or any similar frame with a list of SMILES codes in column ['Smiles']
        drop_attachments: boolean, if False, retains placeholder \* at points of attachment,
            if True, removes \* for fragments with more than one atom
        uni_smi_type: include atom types in determination of unique
            fragments. If false, only determine unique by SMILES

    Returns:
        pandas data frame with columns 'Smiles', 'count' and 'Molecule',
        containing the Smiles string, the number of times the Smiles was in frag_frame,
        and rdkit.Chem.Molecule object

    Note:
        if drop_attachments=False, similar fragments with different number/positions of
        attachments will not count as being the same.
        e.g. ortho-attached aromatics would not match with meta or para attached aromatics

        If you've ran this previously with uni_smi_type=True, running on the output frame
        (or other frame derived from such frame)
        with uni_smi_type=False will collapse the output uniques determined by SMILE only

    """
    # Change column indices to dict
    # pylint:disable=too-many-locals
    # pylint:disable=too-many-branches
    # pylint:disable=too-many-statements
    col_names = list(frag_frame.columns)
    # if frag_frame already has xyz coordinates keep those and don't use others
    # typically this will be if a mol and/or cml file was provided in frag_frame construction
    if "xyz" in col_names:
        xyz_inc = True
        xyz_list = frag_frame["xyz"]
    else:
        xyz_inc = False
    smile_list = frag_frame["Smiles"]
    if "Atoms" in col_names:
        atom_list = frag_frame["Atoms"]
        atoms_inc = True
    else:
        atoms_inc = False
    if "Labels" in col_names:
        labels_list = frag_frame["Labels"]
        labels_inc = True
    else:
        labels_inc = False
    if "Parent" in col_names:
        parent_list = frag_frame["Parent"]
        parent_inc = True
    else:
        parent_inc = False
    if "atom_types" in col_names:
        type_list = frag_frame["atom_types"]
        type_inc = True
    else:
        type_inc = False
    if "count" in col_names:
        count_list = frag_frame["count"]
        count_inc = True
    else:
        count_inc = False
    # print(type_list)
    no_connect_smile = []
    # Clean smiles - either removing placeholder entirely(drop_attachments True)
    # Or just removing the dummyAtomLabel (drop_attachments False)
    for smile in smile_list:
        if drop_attachments:
            no_connect_smile.append(_drop_smi_attach(smile))
        else:
            t_mol = Chem.MolFromSmiles(re.sub(r"\[[0-9]+\*\]", "*", smile))
            no_connect_smile.append(Chem.MolToSmiles(t_mol))
    # identify unique smiles and count number of times they occur
    # initialize lists to be used making frame
    unique_smiles = []
    unique_xyz = []
    unique_atoms = []
    unique_labels = []
    unique_parents = []
    unique_smiles_counts = []
    unique_types = []
    counter_list = []
    # Identify unique smiles, counting every occurence and adding xyz if included
    if uni_smi_type:
        ty_tup = list(map(tuple, frag_frame["atom_types"]))
        smi_ty = list(zip(no_connect_smile, ty_tup))
        for i, val in enumerate(smi_ty):
            smile = val[0]
            at_tys = val[1]
            if smile not in unique_smiles:
                unique_smiles.append(smile)
                if xyz_inc:
                    unique_xyz.append(xyz_list[i])
                if atoms_inc:
                    unique_atoms.append(atom_list[i])
                if labels_inc:
                    unique_labels.append(labels_list[i])
                if parent_inc:
                    unique_parents.append(parent_list[i])
                if type_inc:
                    unique_types.append(type_list[i])
                    counter_list.append(Counter(at_tys))
                if count_inc:
                    unique_smiles_counts.append(count_list[i])
                else:
                    unique_smiles_counts.append(1)
            else:
                at_ty_count = Counter(at_tys)
                if at_ty_count in counter_list:
                    smi_ix = counter_list.index(at_ty_count)
                    if count_inc:
                        unique_smiles_counts[smi_ix] += count_list[i]
                    else:
                        unique_smiles_counts[smi_ix] += 1
                else:
                    unique_smiles.append(smile)
                    if xyz_inc:
                        unique_xyz.append(xyz_list[i])
                    if atoms_inc:
                        unique_atoms.append(atom_list[i])
                    if labels_inc:
                        unique_labels.append(labels_list[i])
                    if parent_inc:
                        unique_parents.append(parent_list[i])
                    # if type_inc:
                    #     unique_types.append(labels_list[i])
                    counter_list.append(Counter(at_tys))
                    unique_types.append(type_list[i])
                    if count_inc:
                        unique_smiles_counts.append(count_list[i])
                    else:
                        unique_smiles_counts.append(1)
    else:
        for i, smile in enumerate(no_connect_smile):
            if smile not in unique_smiles:
                unique_smiles.append(smile)
                if xyz_inc:
                    unique_xyz.append(xyz_list[i])
                if atoms_inc:
                    unique_atoms.append(atom_list[i])
                if labels_inc:
                    unique_labels.append(labels_list[i])
                if parent_inc:
                    unique_parents.append(parent_list[i])
                if type_inc:
                    unique_types.append(type_list[i])
                if count_inc:
                    unique_smiles_counts.append(count_list[i])
                else:
                    unique_smiles_counts.append(1)
            else:
                smi_ix = unique_smiles.index(smile)
                if count_inc:
                    unique_smiles_counts[smi_ix] += count_list[i]
                else:
                    unique_smiles_counts[smi_ix] += 1
    # create output frame

    un_frame = _construct_unique_frame(
        uni_smi=unique_smiles,
        uni_smi_count=unique_smiles_counts,
        xyz=unique_xyz,
        atoms=unique_atoms,
        parents=unique_parents,
        labels=unique_labels,
        at_types=unique_types,
    )
    # if atoms_inc:
    #     un_frame['Atoms'] = unique_atoms
    # if labels_inc:
    #     un_frame['Labels'] = unique_labels
    # if parent_inc:
    #     un_frame['Parent'] = unique_parents
    # if type_inc:
    #     un_frame['atom_types'] = unique_types
    return un_frame


def _find_neigh_notin_frag(mol, at_list):
    # for atoms with one attachment point
    out_nbr = 0
    for idx in at_list:
        atom = mol.GetAtomWithIdx(idx - 1)
        nbrs = atom.GetNeighbors()
        for nbr in nbrs:
            nbr_idx = nbr.GetIdx() + 1
            if nbr_idx not in at_list:
                out_nbr = nbr_idx
                out_at = atom.GetIdx() + 1
                break
        if out_nbr:
            break
    return [out_at, out_nbr]


def _find_neigh_xyz(frag_frame, neigh_idx):
    """Given frag_frame and index of neighbor atom, find its xyz coordinates"""
    # column index
    atoms_idx = list(frag_frame.columns).index("Atoms")
    # determine which row contains the neighbor atom
    neigh_bool = list(frag_frame.apply(lambda row: neigh_idx in row[atoms_idx], axis=1))
    # print(neigh_bool)
    neigh_row = neigh_bool.index(True)
    list_xyz = frag_frame["xyz"][neigh_row]
    # find where in list neighbor atom is, return xyz at that index
    neigh_list_idx = frag_frame["Atoms"][neigh_row].index(neigh_idx)
    neigh_xyz = list_xyz[neigh_list_idx]
    return neigh_xyz


def _move_along_bond(at_xyz, neigh_xyz, at_symb):
    """Returns xyz coordinates of neigh_xyz moved along bond to H bond length"""
    at_np = np.array(at_xyz)
    nb_np = np.array(neigh_xyz)
    # vector for bond to move along
    bond = at_np - nb_np
    # bond length to find
    target_length = _H_BOND_LENGTHS[at_symb]
    steps = np.linspace(0.0, 1.0, 100)
    # step along bond starting at neigh_xyz in direction of at_xyz.
    # stop when bond length is target
    for s in steps:
        new_xyz = neigh_xyz + s * bond
        if abs(_get_dist(new_xyz, at_np) - target_length) < 0.01:
            end_xyz = new_xyz
    # return target point
    return end_xyz


def _get_dist(point_a, point_b):
    """Return distance btw two points"""
    return math.sqrt(
        (point_a[0] - point_b[0]) ** 2
        + (point_a[1] - point_b[1]) ** 2
        + (point_a[2] - point_b[2]) ** 2
    )


def _find_H_xyz(mol, at_list, xyz_list, frag_frame):
    """Finds xyz of hydrogen atom that would be connected to a fragment

    Args:
        mol: Chem.Mol object
        at_list: list[int] - list of atoms in the fragment
            Note: these start at 1, but those in molecule start at 0.
            To convert add/subtract 1
        xyz_list: list of xyz_coordinates of atoms in fragment
        frag_frame: full parent frag_frame WITHOUT filtering by number of attachments

    Returns:
        list[float]: xyz coordinates where H should be placed
    """
    # find attached atom index and neighbor index in molecule
    at_idx, neigh_idx = _find_neigh_notin_frag(mol, at_list)
    # get attached atom symbol
    symb = mol.GetAtomWithIdx(at_idx - 1).GetSymbol()
    # index of the attached atom in at_list, which is the same as xyz_list
    list_idx = at_list.index(at_idx)
    # attached atom xyz
    at_xyz = xyz_list[list_idx]
    # neighbor atom xyz
    neigh_xyz = _find_neigh_xyz(frag_frame, neigh_idx)
    # move neighbor atom xyz along bond to H bond length (Gaussview defaults)
    h_xyz = _move_along_bond(at_xyz, neigh_xyz, symb)
    return h_xyz


def _clean_molecule_name(smile):
    """Removes symbols in smile code unfit for file name"""
    smile = smile.replace("-", "Neg")
    smile = smile.replace("[", "-")
    smile = smile.replace("]", "-")
    smile = smile.replace("(", "-")
    smile = smile.replace(")", "-")
    smile = smile.replace("#", "t")
    smile = smile.replace("=", "d")
    smile = smile.replace("+", "Pos")
    smile = smile.replace("*", "Att")
    smile = smile.replace("@", "")
    return smile


# def generate_fragment_structures(
#     inp: str,
#     keep_only_children: bool = True,
#     bb_patt: str = "[$([C;X4;!R]):1]-[$([R,!$([C;X4]);!#0;!#9;!#17;!#35;!#1]):2]",
#     input_type="smile",
#     cml_file="",
#     include_parent=True,
# ):
#     """generate fragments and return a dictionary of fragments"""
#     # pylint:disable=too-many-arguments
#     frag_frame = identify_connected_fragments(
#         inp=inp,
#         keep_only_children=keep_only_children,
#         bb_patt=bb_patt,
#         input_type=input_type,
#         cml_file=cml_file,
#         include_parent=include_parent,
#     )
#     mol = frag_frame.at[0, "Parent"]
#     frag_dict = output_ifc_dict(mol, frag_frame)
#     return frag_dict


# show


# def output_cgis_dicts(cgis_frame):
#     """generate fragments and return a dictionary of fragments"""
#     # should have parent col
#     on_at_frame = pd.DataFrame(cgis_frame[cgis_frame["numAttachments"] == 1])
#     col_names = list(on_at_frame.columns)
#     xyz_idx = col_names.index("xyz")
#     atoms_idx = col_names.index("Atoms")
#     parent_idx = col_names.index("Parent")
#     on_at_frame["H_xyz"] = on_at_frame.apply(
#         lambda row: _find_H_xyz(
#             row[parent_idx], row[atoms_idx], row[xyz_idx], cgis_frame
#         ),
#         axis=1,
#     )
#     on_at_frame["at_idx"] = on_at_frame.apply(
#         lambda row: _find_at_idx(row[parent_idx], row[atoms_idx]), axis=1
#     )
#     out_dict = {
#         re.sub(
#             r"\[[0-9]+\*\]", "*", on_at_frame.at[i, "Smiles"]
#         ): _write_frag_structure(
#             frag_mol=on_at_frame.at[i, "Molecule"],
#             xyz_list=on_at_frame.at[i, "xyz"],
#             symb_list=on_at_frame.at[i, "Labels"],
#             h_xyz=on_at_frame.at[i, "H_xyz"],
#             at_idx=on_at_frame.at[i, "at_idx"],
#             atom_types=frag_frame.at[i, "atom_types"],
#         )
#         for i in range(0, nrow)
#     }
#     return out_dict


def output_ifc_dict(mol, frag_frame: pd.DataFrame, done_smi: list[str]):
    """generate a dictionary containing identify_connected_fragment information

    Only new fragments are included. Previously parsed fragments are listed in done_smi.

    Args:
        mol: rdkit molecule object that was fragmented
        frag_frame: identify_connected_fragments frame generated for mol
        done_smi: list of fragments which have been identified already

    Returns:
        list containing dict for fragments and done_smi lengthened by the number of fragments done

    Note:
        mainly for use in generating information for unique fragments in
        an AiiDA workflow

    """
    on_at_frame = pd.DataFrame(frag_frame[frag_frame["numAttachments"] == 1])
    if on_at_frame.empty:
        return None, done_smi
    col_names = list(on_at_frame.columns)
    # Find indices of relevant columns
    xyz_idx = col_names.index("xyz")
    atoms_idx = col_names.index("Atoms")

    # Find H xyz position and index of atom bonded to H
    on_at_frame["H_xyz"] = on_at_frame.apply(
        lambda row: _find_H_xyz(mol, row[atoms_idx], row[xyz_idx], frag_frame), axis=1
    )
    on_at_frame["at_idx"] = on_at_frame.apply(
        lambda row: _find_at_idx(mol, row[atoms_idx]), axis=1
    )

    nrow = on_at_frame.shape[0]
    on_at_frame = on_at_frame.reset_index(drop=True)
    # print(on_at_frame)
    # print(_clean_molecule_name(on_at_frame.at[0,'Smiles']))
    # print(on_at_frame.at[0, "Molecule"])
    # print(on_at_frame.at[0, "Labels"])
    # print(on_at_frame.at[0, "xyz"])
    # print(on_at_frame.at[0, "H_xyz"])
    # print(on_at_frame.at[0, "at_idx"])
    # print(on_at_frame.at[0, "atom_types"])

    out_dict = {}
    for i in range(0, nrow):
        smi = on_at_frame.at[i, "Smiles"]
        if smi not in done_smi:
            done_smi.append(smi)
            key = re.sub(r"\[[0-9]+\*\]", "*", smi)
            out_dict[key] = _write_frag_structure(
                frag_mol=on_at_frame.at[i, "Molecule"],
                xyz_list=on_at_frame.at[i, "xyz"],
                symb_list=on_at_frame.at[i, "Labels"],
                h_xyz=on_at_frame.at[i, "H_xyz"],
                at_idx=on_at_frame.at[i, "at_idx"],
                atom_types=frag_frame.at[i, "atom_types"],
            )
    return [out_dict, done_smi]


def _write_frag_structure(frag_mol, xyz_list, symb_list, h_xyz, at_idx, atom_types):
    """get output structure"""
    # pylint:disable=too-many-arguments
    # print("in function")
    num_atoms = len(symb_list)
    # smile = re.sub('\[[0-9]+\*\]', '*', Chem.MolToSmiles(frag_mol,canonical=False))
    charge = Chem.GetFormalCharge(frag_mol)
    out_xyz = [xyz_list[at_idx], h_xyz]
    for i in range(num_atoms):
        if i != at_idx:
            out_xyz.append(xyz_list[i])
    # geom_frame = pd.DataFrame(out_xyz,columns=['x','y','z'])
    symb_list.insert(1, "H")
    return {
        "geom": out_xyz,
        "charge": charge,
        "atom_types": atom_types,
        "atom_symbols": symb_list,
    }


def output_ifc_gjf(
    mol,
    frag_frame,
    esm="wb97xd",
    basis_set="aug-cc-pvtz",
    wfx=True,
    n_procs=4,
    mem="3200MB",
    multiplicity=1,
):
    """Takes a fragmented molecule and outputs gjf files of the fragments with one
    attachment point.

    Hydrogen is added in place of the connection to the rest of
    the molecule for the fragment

    Args:
        mol: Chem.Mol object for which fragmentation was performed
        frag_frame: output from either count_uniques or identify_connected_fragments
        esm: str, electronic structure method to include in gjf
        basis_set: str, basis set to include in gjf
        wfx: Boolean, if True add output=wfx to gjf file
        n_procs: int, >=0. if >0, add number of processors to be used to gjf
        mem: str, format "nMB" or "nGB", memory to be used in gjf
        multiplicity: int, defaults to 1. Multiplicity of molecule

    Returns:
        Creates gjf files in working directory for each fragment in frag_frame

    Note:
        H position is set by taking the atom the fragment is bonded two, replacing it with H
        and moving that closer to the C until it reaches the default distance

        Default distances taken from Gaussview "clean" C-H, C-O, etc bond lengths

    """
    on_at_frame = pd.DataFrame(frag_frame[frag_frame["numAttachments"] == 1])
    col_names = list(on_at_frame.columns)
    # Find indices of relevant columns
    xyz_idx = col_names.index("xyz")
    atoms_idx = col_names.index("Atoms")
    labels_idx = col_names.index("Labels")
    mol_idx = col_names.index("Molecule")
    # Find H xyz position and index of atom bonded to H
    on_at_frame["H_xyz"] = on_at_frame.apply(
        lambda row: _find_H_xyz(mol, row[atoms_idx], row[xyz_idx], frag_frame), axis=1
    )
    on_at_frame["at_idx"] = on_at_frame.apply(
        lambda row: _find_at_idx(mol, row[atoms_idx]), axis=1
    )
    # hxyz = len(col_names)
    # atidx = len(col_names) + 1
    # print(on_at_frame)
    # write gjfs
    on_at_frame.apply(
        lambda row: _write_frag_gjf(
            frag_mol=row[mol_idx],
            xyz_list=row[xyz_idx],
            symb_list=row[labels_idx],
            h_xyz=row[len(col_names)],
            at_idx=row[len(col_names) + 1],
            esm=esm,
            basis_set=basis_set,
            wfx=wfx,
            n_procs=n_procs,
            mem=mem,
            multiplicity=multiplicity,
        ),
        axis=1,
    )


def _clean_basis(basis_set):
    basis_set = basis_set.replace("(", "")
    basis_set = basis_set.replace(")", "")
    basis_set = basis_set.replace(",", "")
    basis_set = basis_set.replace("+", "p")
    return basis_set


def _write_frag_gjf(
    frag_mol,
    xyz_list,
    symb_list,
    h_xyz,
    at_idx,
    esm="wb97xd",
    basis_set="aug-cc-pvtz",
    wfx=True,
    n_procs=4,
    mem="3200MB",
    multiplicity=1,
):
    """Write the gjf file for a fragment.
    Args:
        frag_mol: Chem.Mol object
        xyz_list: list of xyz_coords of list
        symb_list: list of atomic symbols of fragment
        h_xyz: xyz coords of hydrogen attached to fragment
        at_idx: index of attached atom in fragment
        esm: str, electronic structure method to include in gjf
        basis_set: str, basis set to include in gjf
        wfx: Boolean, if True add output=wfx to gjf file
        n_procs: int, >=0. if >0, add number of processors to be used to gjf
        mem: str, format "nMB" or "nGB", memory to be used in gjf
        multiplicity: int, defaults to 1. Multiplicity of molecule

    Default template:
        %chk={filename}.chk
        %nprocs=4
        %mem=3200MB
        # esm/basis_set opt freq output=wfx

        smile

        charge mult
        {xyz}

        {filename}.wfx

    """
    # pylint:disable=too-many-locals
    num_atoms = len(symb_list)
    smile = re.sub(r"\[[0-9]+\*\]", "*", Chem.MolToSmiles(frag_mol, canonical=False))
    charge = Chem.GetFormalCharge(frag_mol)
    molecule_name = _clean_molecule_name(smile)
    # print(at_idx)
    # print(xyz_list[at_idx])
    # print(h_xyz)
    # build xyz of molecule
    out_xyz = [xyz_list[at_idx], h_xyz]
    for i in range(num_atoms):
        if i != at_idx:
            out_xyz.append(xyz_list[i])
    geom_frame = pd.DataFrame(out_xyz, columns=["x", "y", "z"])
    symb_list.insert(1, "H")
    geom_frame["Atom"] = symb_list
    geom_frame = geom_frame[["Atom", "x", "y", "z"]]
    # create file name
    clean_basis = _clean_basis(basis_set)
    new_file_name = "SubH" + "_" + molecule_name + "_" + esm + "_" + clean_basis
    if os.path.exists(new_file_name + ".gjf"):
        # print('deleting')
        os.remove(new_file_name + ".gjf")
    # write file
    with open(new_file_name + ".gjf", "a", encoding="utf-8") as f:
        f.write(f"%chk={new_file_name}" + ".chk\n")
        if n_procs:
            f.write(f"%nprocs={n_procs}\n")
        if mem:
            f.write(f"%mem={mem}\n")
        if wfx:
            f.write(f"#p {esm}/{basis_set} opt freq output=wfx\n")
        else:
            f.write(f"#p {esm}/{basis_set} opt freq\n")
        f.write("\n")
        f.write(f"{smile}")
        f.write("\n\n")
        f.write(f"{charge} {multiplicity}\n")
        dfAsString = geom_frame.to_string(header=False, index=False)
        f.write(dfAsString)
        f.write("\n\n")
        if wfx:
            f.write(new_file_name + ".wfx\n\n\n")
        else:
            f.write("\n\n\n")


def _find_at_idx(mol, at_list):
    """Returns index of atom connected to the remainder of the molecule in the list
    of fragment atom numbers

    Args:
        mol: Chem.Mol object
        at_list: list[int] of atoms in molecule belonging to fragment being studied
            Note: these indices start at 1, while those in the molecule start at 0

    Returns:
        int of index of connected atom in at_list. See examples for worked explanation

    Example: a fragment is defined by atoms [1,3,5,6,8]
    Atom 3 is where the remainder of the molecule attaches to the fragment
    The function returns the index of 3 in the atom list.
    In this case, it would return 1
    """
    # find
    at_idx = _find_neigh_notin_frag(mol, at_list)[0]
    list_idx = at_list.index(at_idx)
    return list_idx


def _construct_unique_frame(
    uni_smi: list[str],
    uni_smi_count: list[int],
    xyz=[],
    atoms=[],
    parents=[],
    labels=[],
    at_types=[],
) -> pd.DataFrame:
    """given smiles, counts and (optional) xyz coordinates, create frame"""
    # pylint:disable=dangerous-default-value
    uniquefrag_frame = pd.DataFrame(uni_smi, columns=["Smiles"])
    if xyz:
        uniquefrag_frame["xyz"] = xyz
    if atoms:
        uniquefrag_frame["Atoms"] = atoms
    if parents:
        uniquefrag_frame["Parent"] = parents
    if labels:
        uniquefrag_frame["Labels"] = labels
    if at_types:
        uniquefrag_frame["atom_types"] = at_types
    PandasTools.AddMoleculeColumnToFrame(
        uniquefrag_frame, "Smiles", "Molecule", includeFingerprints=True
    )
    uniquefrag_frame["count"] = uni_smi_count
    # if we don't have xyz already add them, from MMFF94 opt
    if not xyz:
        uniquefrag_frame = _add_xyz_coords(uniquefrag_frame)
    # count number placeholders
    uniquefrag_frame = _add_number_attachements(uniquefrag_frame)
    return uniquefrag_frame


def _drop_smi_attach(smile: str):
    """completely remove placeholder if number of non-placeholder in smiles is > 1"""
    mol = Chem.MolFromSmiles(smile)
    non_zero_atoms = 0
    for atom in mol.GetAtoms():
        if atom.GetAtomicNum() != 0:
            non_zero_atoms += 1
    if non_zero_atoms > 1:
        temp = re.sub(r"\[[0-9]+\*\]", "", smile)
        t_mol = Chem.MolFromSmiles(re.sub(r"\(\)", "", temp))
        if t_mol is None:
            t_mol = Chem.MolFromSmiles(re.sub(r"\[[0-9]+\*\]", "*", smile))
            out_smi = Chem.MolToSmiles(t_mol)
            # Warning('Could not construct {smile} without attachments'.format(smile=smile))
        else:
            out_smi = Chem.MolToSmiles(t_mol)
    else:
        out_smi = Chem.MolToSmiles(
            Chem.MolFromSmiles(re.sub(r"\[[0-9]+\*\]", "*", smile))
        )
    return out_smi


def merge_uniques(
    frame1: pd.DataFrame, frame2: pd.DataFrame, uni_smi_ty=True
) -> pd.DataFrame:
    """Given two frames of unique fragments, identify shared unique fragments,
    merge count and frames together.

    Args:
        frame1: a frame output from count_uniques
        frame2: a distinct frame also from count_uniques
        uni_smi_ty: If True, include atom types in determination of unique
            fragments. If false, only determine unique by SMILES

    Returns:
        a frame resulting from the merge of frame1 and frame2.
        All rows that have Smiles that are in frame1 but not frame2(and vice versa) are included
        unmodified
        If a row's SMILES is in both frame1 and frame2, modify the row to update the count of
        that fragment as sum of frame1 and frame2, then include one row.

    Note:
        for best results, SMILES must be canonical so that they can be exactly compared.
        Smiles in frame should be resulting from Chem.MolToSmiles(Chem.MolFromSmiles(smile)) -
        this will create a molecule from the smile, and write the smile back, in canonical form

    Example usage:
        >>> frame1
        Smiles  count
        C       2
        C1CCC1  1

        >>> frame2
        Smiles  count
        C       3
        C1CC1   2

        >>> merge_uniques(frame1,frame2)
        Smiles  count
        C       5
        C1CCC1  1
        C1CC1   2

    """
    if frame1.empty:
        merge_frame = frame2
    elif frame2.empty:
        merge_frame = frame1
    else:
        rows_to_drop = _find_rows_to_drop(frame1, frame2, uni_smi_ty)
        merge_frame = rows_to_drop["merge_frame"]
        drop_frame_1 = frame1.drop(rows_to_drop["drop_rows_1"])
        drop_frame_2 = frame2.drop(rows_to_drop["drop_rows_2"])
        merge_frame = pd.concat([drop_frame_1, drop_frame_2, merge_frame])
        merge_frame.reset_index(inplace=True, drop=True)
        # _add_frame(drop_frame_1,merge_frame)
        # merge_frame = _add_frame(drop_frame_2,merge_frame)
    return merge_frame


def _find_rows_to_drop(frame_a: pd.DataFrame, frame_b: pd.DataFrame, uni_smi_ty=True):
    """for two frames, find rows with same smile, store what rows to drop in each"""
    rows_to_drop_one = []
    rows_to_drop_two = []
    col_names = list(frame_a.columns)
    merge_frame = pd.DataFrame(columns=col_names)
    frame_a_idx = list(frame_a.index)
    frame_b_idx = list(frame_b.index)
    for i, smi in enumerate(frame_a["Smiles"]):
        if smi in list(frame_b["Smiles"]):
            # print(f'i is {i}')
            if uni_smi_ty:
                smi_idx = [
                    x
                    for x in range(0, frame_b.shape[0])
                    if smi == list(frame_b["Smiles"])[x]
                ]
                ctr = [list(map(Counter, frame_b["atom_types"]))[x] for x in smi_idx]
                if list(map(Counter, frame_a["atom_types"]))[i] not in ctr:
                    continue
                j = smi_idx[ctr.index(list(map(Counter, frame_a["atom_types"]))[i])]
            else:
                j = list(frame_b["Smiles"]).index(smi)
            # print(f'j is {j}')
            # print(frame_a.shape)
            cum_count = (
                frame_a.at[frame_a_idx[i], "count"]
                + frame_b.at[frame_b_idx[j], "count"]
            )
            # print('cum_count is {cum_count}')
            merge_frame = pd.concat(
                [
                    merge_frame,
                    pd.DataFrame(
                        [list(frame_a.iloc[frame_a_idx[i]])], columns=col_names
                    ),
                ]
            )
            merge_frame.reset_index(inplace=True, drop=True)
            # print(f'merge frame before updating cumcount {merge_frame}')
            # print(merge_frame.shape[0]-1)
            merge_frame.at[merge_frame.shape[0] - 1, "count"] = cum_count
            # print(f'merge_frame after updating cumcount {merge_frame}')
            # if made_frame == 0:
            #     merge_frame = pd.DataFrame.from_dict({'Smiles':[smi],'count':[cum_count]})
            #     made_frame=1
            # else:
            #     merge_frame.loc[len(merge_frame)] = [smi, cum_count]
            rows_to_drop_one.append(i)
            rows_to_drop_two.append(j)
    # print(merge_frame)
    return {
        "drop_rows_1": rows_to_drop_one,
        "drop_rows_2": rows_to_drop_two,
        "merge_frame": merge_frame,
    }


def count_groups_in_set(
    list_of_inputs: list[str],
    drop_attachments: bool = False,
    input_type: str = "smile",
    bb_patt: str = "[$([C;X4;!R]):1]-[$([R,!$([C;X4]);!#0;!#9;!#17;!#35;!#1]):2]",
    cml_list=None,
    uni_smi_ty: bool = True,
    aiida: bool = False,
) -> pd.DataFrame:
    """Identify unique fragments in molecules defined in the list_of_smiles,
    and count the number of occurences for duplicates.

    Args:
        list_of_smiles: A list, with each element being a SMILES string, e.g. ['CC','C1CCCC1']
        drop_attachments: Boolean for whether or not to drop attachment points from fragments
            if True, will remove all placeholder atoms indicating connectivity
            if False, placeholder atoms will remain
        input_type: smile, xyzfile, cmlfile or molfile, based on elements of lists_of_inputs
        cml_list: defaults empty, but can be a list of cml files corresponding to the molfile inputs
        bb_patt: SMARTS pattern for bonds to break in linkers and side chains. Defaults to breaking
            bonds between nonring carbons with four bonds single bonded to ring atoms or carbons that
            don't have four bonds, and are not H, halide, or placeholder
        uni_smi_type: if True, include atom types in determination of unique
            fragments. If false, only determine unique by SMILES

    Returns:
        an output pd.DataFrame, with columns 'Smiles' for fragment Smiles,
        'count' for number of times each fragment occurs in the list, and
        'Molecule' holding a rdkit.Chem.Molecule object

    Example usage:
        >>> count_groups_in_set(['c1ccc(c(c1)c2ccc(o2)C(=O)N3C[C@H](C4(C3)CC[NH2+]CC4)C(=O)NCCOCCO)F',
        'Cc1nc2ccc(cc2s1)NC(=O)c3cc(ccc3N4CCCC4)S(=O)(=O)N5CCOCC5'],drop_attachments=False)

    """

    out_frame = pd.DataFrame()
    for i, inp in enumerate(list_of_inputs):
        # print(inp)
        if cml_list:
            frame = identify_connected_fragments(
                inp,
                bb_patt=bb_patt,
                input_type=input_type,
                cml_file=cml_list[i],
                include_parent=True,
            )
        else:
            frame = identify_connected_fragments(
                inp, bb_patt=bb_patt, input_type=input_type, include_parent=True
            )
        if frame is not None:
            unique_frame = count_uniques(frame, drop_attachments, uni_smi_type=True)
            if out_frame.empty:
                out_frame = unique_frame
            else:
                out_frame = merge_uniques(out_frame, unique_frame, uni_smi_ty)
    out_frame = out_frame.drop("Molecule", axis=1)
    if not aiida:
        PandasTools.AddMoleculeColumnToFrame(
            out_frame, "Smiles", "Molecule", includeFingerprints=True
        )
    # out_frame = _add_xyz_coords(out_frame)
    # out_frame = _add_number_attachements(out_frame)
    return out_frame



"""
fg_query module

Functions and SMARTs strings for querying specific functional groups from results table of
:attr:`group_decomposition.fragfunctions.identify_connected_fragments`
:attr:`group_decomposition.fragfunctions.count_uniques`, or :attr:`group_decomposition.fragfunctions.count_groups_in_set`
"""
import pandas as pd  # lots of work with data frames
from rdkit import Chem  # pylint:disable=import-error

fg_dict = {
    "vinylic carbon": "[$([CX3]=[CX3])]",
    "allenic carbon": "[$([CX2](=C)=C)]",
    "alkyne": "[$([CX2]#C)]",
    "acyl halide": "[CX3](=[OX1])[F,Cl,Br,I]",
    "aldehyde": "[CX3H1](=O)[#6]",
    "anhydride": "[CX3](=[OX1])[OX2][CX3](=[OX1])",
    "amide": "[NX3][CX3](=[OX1])[#6]",
    "amidinium": "[NX3][CX3]=[NX3+]",
    "carbamic ester": "[NX3][CX3](=[OX1])[OX2H0]",
    "carbamic acid": "[NX3,NX4+][CX3](=[OX1])[OX2H,OX1-]",
    "carboxylate": "[CX3](=O)[O-]",
    "carbonic acid/acid-ester": "[CX3](=[OX1])([OX2])[OX2H,OX1H0-1]",
    "carbonic ester": "C[OX2][CX3](=[OX1])[OX2]C",
    "carboxylic acid": "[CX3](=O)[OX2H1]",
    "cyanamide": "[NX3][CX2]#[NX1]",
    "ester or anhydride": "[#6][CX3](=O)[OX2H0][#6]",
    "ketone": "[#6][CX3](=O)[#6]",
    "ether": "[OD2]([#6])[#6]",
    "primary amine": "[NX3;H2;!$(NC=[!#6]);!$(NC#[!#6])][#6]",
    "enamine": "[NX3][CX3]=[CX3]",
    "secondary amine": "[NX3;H1;!$(NC=O)]",
    "azide": "[$(*-[NX2-]-[NX2+]#[NX1]),$(*-[NX2]=[NX2+]=[NX1-])]",
    "hydrazine": "[NX3][NX3]",
    "hydrazone": "[NX3][NX2]=[*]",
    "substituted imine": "[CX3;$([C]([#6])[#6]),$([CH][#6])]=[NX2][#6]",
    "iminium": "[NX3+]=[CX3]",
    "unsubstituted dicarboximide": "[CX3](=[OX1])[NX3H][CX3](=[OX1])",
    "substituted dicarboximide": "[CX3](=[OX1])[NX3H0]([#6])[CX3](=[OX1])",
    # 'nitrate': '[$([NX3](=[OX1])(=[OX1])O),$([NX3+]([OX1-])(=[OX1])O)]',
    "nitrile": "[NX1]#[CX2]",
    "isonitrile": "[CX1-]#[NX2+]",
    "nitro": "[$([NX3](=O)=O),$([NX3+](=O)[O-])][!#8]",
    "nitroso": "[NX2]=[OX1]",
    "alcohol hydroxyl": "[OX2H][CX4]",
    "enol": "[OX2H][#6X3]=[#6]",
    "phenol": "[OX2H][cX3]:[c]",
    "peroxide": "[OX2,OX1-][OX2,OX1-]",
    "thioester": "S([#6])[CX3](=O)[#6]",
    "thiol": "[#16X2H]",
    "thioamide": "[NX3][CX3]=[SX1]",
    "thioketone": "[#6][CX3](=S)[#6]",
    "thioaldehyde": "[CX3H1](=S)[#6]",
    "monosulfide": "[#16X2H0][!#16]",
    "disulfide": "[#16X2H0][#16X2H0]",
    "sulfinic acid": "[$([#16X3](=[OX1])[OX2H,OX1H0-]),$([#16X3+]([OX1-])[OX2H,OX1H0-])]",
    "sulfonic acid": "[$([#16X4](=[OX1])(=[OX1])([#6])[OX2H,OX1H0-]),$([#16X4+2]([OX1-])([OX1-])([#6])[OX2H,OX1H0-])]",
    "sulfonamide": "[$([#16X4]([NX3])(=[OX1])(=[OX1])[#6]),$([#16X4+2]([NX3])([OX1-])([OX1-])[#6])]",
    "sulfone": "[$([#16X4](=[OX1])(=[OX1])([#6])[#6]),$([#16X4+2]([OX1-])([OX1-])([#6])[#6])]",
    "sulfoxide": "[$([#16X3](=[OX1])([#6])[#6]),$([#16X3+]([OX1-])([#6])[#6])]",
    "sulfate": "[$([#16X4](=[OX1])(=[OX1])([OX2H,OX1H0-])[OX2][#6]),$([#16X4+2]([OX1-])([OX1-])([OX2H,OX1H0-])[OX2][#6])]",
    "sulfuric acid diester": "[$([#16X4](=[OX1])(=[OX1])([OX2][#6])[OX2][#6]),$([#16X4](=[OX1])(=[OX1])([OX2][#6])[OX2][#6])]",  # pylint:disable=line-too-long
    "chlorine": "Cl",
    "fluorine": "F",
    "bromine": "Br",
}


def query_pattern(frag_frame: pd.DataFrame, patt: str) -> int:
    """Determine number of functional groups matching pattern in fragment table

    Args:
        frag_frame: fragment frame from output of :attr:`group_decomposition.fragfunctions.identify_connected_fragments`
            :attr:`group_decomposition.fragfunctions.count_uniques`,
            or :attr:`group_decomposition.fragfunctions.count_groups_in_set`
        patt: SMARTs string to match functional group. See :attr:`group_decomposition.fg_query.fg_dict` for SMARTS
            of common groups

    Returns:
        Number of matches of the functional group queried in the table

    Example Usage:
        >>> frag_frame
        Smiles  Molecule        count
        *C=CC=C <mol object>    3
        *C=C    <mol object>    1
        >>> query_pattern(frag_frame,'[$([CX3]=[CX3])]')
        7

    Note:
        If the column has a count of the number of times a fragment occurs (count column), multiply the number of matches
        in a group by the count of times it occurs

    """
    patt_mol = Chem.MolFromSmarts(patt)
    mol_list = list(frag_frame["Molecule"])
    if "count" in list(frag_frame.columns):
        count_list = list(frag_frame["count"])
        patt_count = [
            count_list[i] * len(mol.GetSubstructMatches(patt_mol))
            for i, mol in enumerate(mol_list)
        ]
    else:
        patt_count = [len(mol.GetSubstructMatches(patt_mol)) for mol in mol_list]
    return sum(patt_count)


def count_fgs(frag_frame: pd.DataFrame, patt_dict: dict) -> pd.DataFrame:
    """Given functional group dictionary with patterns to match, return counts of each in fragment_frame

    Args:
        frag_frame: fragment frame from output of :attr:`group_decomposition.fragfunctions.identify_connected_fragments`
            :attr:`group_decomposition.fragfunctions.count_uniques`,
            or :attr:`group_decomposition.fragfunctions.count_groups_in_set`
        patt_dict: has keys of functional group names and values being a SMARTs string to match

    Returns:
        Frame with one column containing names of functional group and another with count

    Example Usage:
        >>> frag_frame
        Smiles  Molecule        count
        *C=CC=C <mol object>    3
        *C=O    <mol object>    1
        >>> patt_dict
        {
        'aldehyde': '[CX3H1](=O)[#6]',
        'vinylic carbon': '[$([CX3]=[CX3])]'
        }
        >>> count_fgs(frag_frame,patt_dict)
        group           count
        aldehye         1
        vinylic carbon  6

    Note:
        If the column has a count of the number of times a fragment occurs (count column), multiply the number of matches
        in a group by the count of times it occurs

    """
    return pd.DataFrame(
        {
            "group": [list(patt_dict.keys())],
            "count": [query_pattern(frag_frame, value) for value in patt_dict.values()],
        }
    )
